/*
 * Purrpackage - http://purrpackage.sourceforge.net
 * Copyright (C) 2010 John Charles Roth
 *
 * Purrpackage is free software, licensed under the terms of the GNU GPL 
 * Version 2 or, at your option, any later version. You should have 
 * received a copy of the license with this file. See the above web address
 * for more information, or contact the Free Software Foundation, Boston, MA. 
 * It is distributed WITHOUT WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
package net.sourceforge.purrpackage.recording;

import java.io.Closeable;
import java.util.ArrayList;
import java.util.List;

/**
 * Enables better closing of file handles and similar things, as follows:
 * 
 * <pre>
 * Closer c = new Closer();
 * // register(c) returns the argument unmodified, for convenience. 
 * FileInputStream is = c.register( new FileInputStream( aFile ) );
 * c.register( out );
 * try {
 *    // do stuff with is and os.
 * }
 * catch( IOException e ) {
 *    // We try to close, but eat exceptions so
 *    // we don't hide this one.
 *    c.doNotThrow( ); 
 *    throw e; // Do not carelessly omit.
 * }
 * finally {
 *     // Try to close both is and os;
 *     // If either fails, throw a runtime exception
 *     // unless c.doNotThrow() has been called.
 *     c.closeAll();
 * }
 * </pre>
 */
public class Closer {

	List<Closeable> items = new ArrayList<Closeable>();

	boolean doNotThrow = false;

	public <T extends Closeable> T register( T item ) {
		items.add( item );
		return item;
	}
	
	public void doNotThrow() {
		doNotThrow = true;
	}

	public void closeAll() {
		Exception last = null;
		int failures = 0;
		for( Closeable c : items ) {
			try {
				c.close();
			}
			catch( Exception e ) {
				failures++;
				last = e;
				try {
					logExceptionWhileClosing( c, e );
				}
				catch( RuntimeException re ) {
				}
			}
		}
		if ( ! doNotThrow ) {
			if ( failures > 0 ) {
				throw new RuntimeException( "Failed to close " + failures + " of " + items.size() + " streams.", last );
			}
		}
	}

	/**
	 * Hook for your logging system, if you wish. Exceptions thrown from here
	 * will be ignored.
	 */
	
	protected void logExceptionWhileClosing( Closeable c, Exception e ) {
	}
}
