/*
 * Purrpackage - http://purrpackage.sourceforge.net
 * Copyright (C) 2010 John Charles Roth
 *
 * Purrpackage is free software, licensed under the terms of the GNU GPL 
 * Version 2 or, at your option, any later version. You should have 
 * received a copy of the license with this file. See the above web address
 * for more information, or contact the Free Software Foundation, Boston, MA. 
 * It is distributed WITHOUT WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
package net.sourceforge.purrpackage.reporting;
import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.jar.JarInputStream;
import java.util.zip.GZIPOutputStream;
import java.util.zip.ZipEntry;

import org.apache.commons.codec.binary.Base64OutputStream;

import net.sourceforge.cobertura.coveragedata.SourceFileData;
import net.sourceforge.cobertura.reporting.html.JavaToHtml;
import net.sourceforge.purrpackage.model.LineDataPair;
import net.sourceforge.purrpackage.model.SourceFileDataPair;
import net.sourceforge.purrpackage.parsing.PerPackageCoverageRunSummary;
import net.sourceforge.purrpackage.reporting.json.MainJsonFormatter;
import net.sourceforge.purrpackage.reporting.json.SourceFileJsonFormatter;
import net.sourceforge.purrpackage.reporting.json.org.json.JSONException;
import net.sourceforge.purrpackage.reporting.json.org.json.JSONObject;


public class JsonReport {

	ReportSupport support;
	SourceReader reader;
	MainJsonFormatter formatter;
	SourceFileJsonFormatter sourceFormatter = new SourceFileJsonFormatter();
	File jsonDir;
	JSONObject tree;

	List<String> reportErrors = new ArrayList<String>();
	

	/**
	 * Create a coverage report
	 */
	public JsonReport( ReportSupport support ) {
		this.support = support;
		reader = new SourceReader( support.getFinder(), support.getEncoding() );
		formatter = new MainJsonFormatter( support.getComplexity(), reader );
	}
	
	public void report() throws Exception {
		prepDestinationDir();
		tree = generateBasicJsonTree();
		saveTree();
		generateSourceFiles();
		extractViews();
		manageCoveragePolicy( );
		if ( runtimeBuildScriptExists() ) {
			runRuntimeBuildScript();
		}
	}

	void prepDestinationDir() throws IOException {
		File destDir = support.getDestinationDir();
		if (destDir.exists()) {
			fileHelper.rmRec( destDir );
		}
		destDir.mkdirs();
		jsonDir = new File(destDir, "jsonCoverageData");
		jsonDir.mkdirs();
	
	}

	File runtimeScriptFile() {
		return new File(support.getBuildTimeReportScriptPath());
	}
	
	boolean runtimeBuildScriptExists() {
		boolean result = false;
		if ( support.getBuildTimeReportScriptPath() != null) {
			File script = runtimeScriptFile();
			if (script.exists()) {
				result = true;
			}
		}
		return result;
	}
	
	void runRuntimeBuildScript() throws Exception {
		BuildTimeJavascriptReporter btr = makeBuildTimeJavscriptReporter(
				support.getDestinationDir(), runtimeScriptFile(), reportErrors);
		btr.report();		
	}
	
	BuildTimeJavascriptReporter makeBuildTimeJavscriptReporter( File f, File script, Collection<String> c ) throws Exception {
		return new BuildTimeJavascriptReporter( f, script, c );
	}
	
	public List<String> getReportErrors() {
		return reportErrors;
	}

	FileHelper fileHelper = new FileHelper();
	
	void manageCoveragePolicy( ) throws IOException {
		fileHelper.copyFileOrFileInDirectoryIfExists( support.getCoveragePolicyPath(), "coveragePolicy.js", jsonDir );
	}
	
	InputStream getJarredViewsInputStream() throws IOException {
		return JsonReport.class.getResourceAsStream("/views.jar");
	}
		
	void extractViews() throws IOException {
		File destinationDir = support.getDestinationDir();
		destinationDir.mkdirs();
		InputStream is = getJarredViewsInputStream();
		JarInputStream jis = new JarInputStream(is);
		ZipEntry x = jis.getNextEntry();
		while (x != null) {
			if (x.isDirectory()) {
				File f = new File( destinationDir, x.getName().substring( 0, x.getName().length() -1 ));
				f.mkdirs();
			} else {
				File f = new File(destinationDir, x.getName() );
				if (!f.exists()) {
					f.createNewFile();
				}
				BufferedOutputStream fos = new BufferedOutputStream(
						new FileOutputStream(f));
				int c = jis.read();
				while (c != -1) {
					fos.write((byte) c);
					c = jis.read();
				}
				fos.flush();
				fos.close();
			}
			x = jis.getNextEntry();
		}
		jis.close();
	}
	
	JSONObject generateBasicJsonTree() throws Exception {
		PerPackageCoverageRunSummary sum = new PerPackageCoverageRunSummary( support.getData(), support.getCoverageDirectoryReader(), 
				support.getCalculator() );
		JSONObject jo = formatter.format( sum );
		return jo;
	}
	
	void saveTree() throws Exception {
		writeCompressedJsonToFile( tree, new File( jsonDir, "coverageData.json.gzuu" ) );
	}

	// Gzip and base64 encode the string for jo
	void writeCompressedJsonToFile( JSONObject jo, File f ) throws IOException, JSONException {
		OutputStream os = new GZIPOutputStream( new Base64OutputStream( new FileOutputStream( f )));
		try {
			OutputStreamWriter w = new OutputStreamWriter( os );
			BufferedWriter b = new BufferedWriter( w );
			jo.write( b );
			b.flush();
		}
		finally {
			os.close();
		}	
	}
	
	void generateSourceFiles() throws Exception {
		int failures = 0;
		int total = 0;
		String lastFailure = null;
		Exception lastException = null;
		for ( Object key : support.getData().getGrandchildKeys() ) {
			total++;
			SourceFileDataPair sourceFileData = (SourceFileDataPair) support.getData().getVanillaDescendent( key );
			try {
				generateSourceFile(sourceFileData);
			}
			catch (Exception e) {
				failures++;
				lastFailure = sourceFileData.getData().getName();
				lastException = e;
			}
		}
		if ( failures > 0 ) {
			StringBuilder msg = new StringBuilder();
			msg.append( "Failed to generate " ).append( failures ).append( " of " ).append( total ).append( " source files: " );
			msg.append( lastFailure ).append( ": " ).append( lastException.toString() );
			reportErrors.add( msg.toString() );
		}
	}

	void generateSourceFile(SourceFileDataPair sourceFileData)
			throws Exception 	{
		String filename = sourceFileData.getData().getNormalizedName() + ".json.gzuu";
		File file = new File(jsonDir, filename);
		JSONObject jo = generateSourceFileObject( sourceFileData );
		writeCompressedJsonToFile( jo, file );
	}

	boolean lacksInstrumentationInfo( SourceFileData sfd ) {
		return !sfd.containsInstrumentationInfo();
	}
	
	JSONObject generateSourceFileObject(final SourceFileDataPair sourceFileDataPair)
			throws Exception {
		final JSONObject result = new JSONObject();
		final JSONObject lineData = new JSONObject();
		result.put( "name", sourceFileDataPair.getName() );
		result.put( "package", sourceFileDataPair.getParent().getName() );
		result.put( "lines", lineData );		
		JSONObject sd = sourceFormatter.format( sourceFileDataPair );
		result.put( "coverage", sd);
		result.put( "issue", "" );
		if ( lacksInstrumentationInfo( sourceFileDataPair.getData() ) ) {
			result.put( "issue", "No instrumentation information for " + sourceFileDataPair.getData().getName() );
			return result;
		}

		final AtomicInteger lineNumber = new AtomicInteger( 1 ); // don't care about concurrency, just handy holder
		result.put( "lineCount", 0 );
		try {
			reader.read( sourceFileDataPair.getData().getName(), new SourceReader.Callback() {
				
				@Override
				public void lineRead(String line) throws Exception {
					lineData.put(String.valueOf(lineNumber), generateRowObject(
							sourceFileDataPair, line, lineNumber.intValue()));
					result.put( "lineCount", lineNumber.intValue() );
					lineNumber.incrementAndGet();
				}
			}
				);
			sd.put( "sourceFileLineCount", lineNumber.get() );
		}
		catch ( Exception e ) {
			result.put( "issue", e.getMessage() );
		}

		return result;
	}
	
	JSONObject generateRowObject( SourceFileDataPair sfdp, String line, int lineNumber ) throws Exception {
		JSONObject result = new JSONObject();
		result.put( "number", lineNumber );
		result.put( "html", javaToHtml.process( line ) );
		if (sfdp.getData().isValidSourceLineNumber(lineNumber)) {
			LineDataPair ld = sfdp.getLine( lineNumber );
			formatLineDataPair( ld, result );
		}
		return result;
	}
	
	void formatLineDataPair( LineDataPair ldp, JSONObject target ) throws Exception {
		if ( ldp != null ) {
			target.put( "coverage", sourceFormatter.format( ldp ) );
		}
	}

	JavaToHtml javaToHtml = new JavaToHtml();
}
